/*
 * $RCSfile: haval.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 *
 * $Log: haval.c,v $
 * Revision 1.7  1995/07/18 03:07:30  spatsch
 * Changed km entries.
 *
 * Revision 1.6  1995/06/08  16:40:20  spatsch
 * fixed warnings.
 *
 * Revision 1.5  1995/06/07  21:37:39  spatsch
 * changed secret handling.
 *
 * Revision 1.4  1995/04/25  16:49:08  spatsch
 * > Moved localsaid used by ipsec in 3rd participnat.
 *
 * Revision 1.3  1995/04/13  20:13:23  spatsch
 * Added IPSEC support.
 *
 *
 */


/* The HAVAL protocol pushes a 32byte cryptographic checksum onto the front
   of a message.  On the incoming side, it pops a 32byte checksum, and
   compares it to the checksum computed for the rest of the message.
   The code for MD5 is directly from the RSA distribution.
*/


#include "xkernel.h"
#include "ip.h"                /* For IP pseudoheader fixup. */
#include "haval_i.h"
#include "havalhash.h"
#include "km.h"
#include "hash.h"
#include "saidmgr.h"

extern int tracehavalp; int tracehavalp;

/*  Byte order note:  All data is grouped in the order ABCD for HAVAL processing. */


#ifdef __STDC__

static XObj		havalCreateSessn( XObj, XObj, XObj, XObj, Part *);
static void		havalSessnInit( XObj );
static int		havalControlProtl( XObj, int, char *, int );
static xkern_return_t	havalDemux( XObj, XObj, Msg * );
static XObj		havalOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	havalOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	havalOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	havalOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t	havalClose( XObj );
static xmsg_handle_t	havalPush( XObj, Msg * );
static xkern_return_t	havalPop( XObj, XObj, Msg *, void * );
static void 		havalStore( void *, char *, long, void * );
static int		havalControlSessn( XObj, int, char *, int );
static int		havalchunk( char *, long, HAVAL_CTX * );
static long 		havalcopy(void *, char*, long, void * );
static long 		nullop(void *, char*, long, void * );
int			xk_haval(unsigned char [], unsigned char *, int, int, unsigned char []);
void haval_init( XObj );
static int		xCopyParts(Part **, char **, Part *, XObj);
static void		xSimpleCopyParts(Part **, Part *);

#else

static XObj		havalCreateSessn();
static void		havalSessnInit();
static int		havalControlProtl();
static xkern_return_t	havalDemux();
static XObj		havalOpen();
static xkern_return_t	havalOpenEnable();
static xkern_return_t	havalOpenDisable();
static xkern_return_t	havalOpenDone();
static xkern_return_t	havalClose();
static xmsg_handle_t	havalPush();
static xkern_return_t	havalPop();
static long 		havalStore();
static long 		havalcopy();
static long 		nullop();
static int		havalControlSessn();
int			xk_haval();
static int		xCopyParts();
static void		xSimpleCopyParts();

#endif /* __STDC__ */

static XObj errobj;

static xkern_return_t romfindpartnum(XObj, char **, int, int, void *);
static xkern_return_t romuseipsec(XObj, char **, int, int, void *);
static xkern_return_t romfindsecretappend(XObj, char **, int, int, void *);
static xkern_return_t romfindtrailer(XObj, char **, int, int, void *);

static XObjRomOpt rmopts[4] = {{ "numparts", 1, romfindpartnum },
			       { "useipsec", 1, romuseipsec },
			       { "secretappend", 1, romfindsecretappend },
			       { "trailers", 1, romfindtrailer }};

static xkern_return_t
romfindpartnum(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  if (atoi(ptr[2]) > 1)
    ((PState *)(self->state))->flags |= ALL_PARTS;
  else
    ((PState *)(self->state))->flags &= ~ALL_PARTS;
  return XK_SUCCESS;
}

static xkern_return_t
romuseipsec(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
    ((PState *)(self->state))->flags |= USE_IPSEC;
    return XK_SUCCESS;
}

static xkern_return_t
romfindsecretappend(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  ((PState *)(self->state))->flags |= HAVAL_SECRETAPPEND;
  return XK_SUCCESS;
}

static xkern_return_t
romfindtrailer(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  xTrace0(havalp, TR_GROSS_EVENTS, "haval rom opt turning on trailer");
  ((PState *)(self->state))->flags |= HAVAL_TRAILERS;
  return XK_SUCCESS;
}

void
haval_init( self )
     XObj self;
{
    PState	*ps;
    XObj	llp;

    xTrace0(havalp, TR_ERRORS, "haval_init entered");
    xAssert(xIsProtocol(self));
    
    if ( self->numdown &&
	! xIsProtocol(llp = xGetDown(self, TRANSPORT_PROTOCOL)) )
      { xError("haval_init: haval cannot get lower protocol object");
	/* return; */}
    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->actMap = mapCreate(HAVAL_ACTIVE_MAP_SZ, sizeof(ActiveId));
    ps->pasMap = mapCreate(HAVAL_PASSIVE_MAP_SZ, sizeof(PassiveId));
    ps->flags = 0;

    self->control = havalControlProtl;
    self->open = havalOpen;
    self->openenable = havalOpenEnable;
    self->opendisable = havalOpenDisable;
    self->opendone = havalOpenDone;
    self->close = havalClose;  
    self->demux = havalDemux;
    bzero((char *)&errobj, sizeof(XObj));
    findXObjRomOpts(self, rmopts, 4, 0);
    xTrace0(havalp, TR_EVENTS, "haval_init exiting");
}

static int
xCopyParts(q, qbuf, p, lls)
     Part **q;
     char **qbuf;
     Part *p;
     XObj lls;
{
  char *ebuf;
  int	len = 500;

  ebuf = xMalloc(500);
  if (lls != ERR_XOBJ) {
    if (xControl(lls, GETPARTICIPANTS, ebuf, len)<0)
      return 0;
  }
  else
    partExternalize(p, ebuf, &len);
  *q = (Part *)xMalloc(partExtLen(ebuf)*sizeof(Part));
  partInternalize(*q, ebuf);
  *qbuf = ebuf;
  return partLen(((Part *)(*q)));
}

static void
xSimpleCopyParts(q, p)
     Part **q;
     Part *p;
{
  int  pl =  partLen(p);

  *q = (Part *)xMalloc(pl*sizeof(Part));
  partInit(*q, pl);
  while (pl--) { (*q)[pl] = p[pl]; }
}

static XObj
havalOpen( self, hlpRcv, hlpType, part )
    XObj	self, hlpRcv, hlpType;
    Part 	*part;
{
    XObj 	havalSessn;
    XObj	lls = errobj;
    PState	*pstate;
    Part	*part2;
    char	*b = 0;

    pstate = (PState *) self->state;
    
    if (self->numdown) {
      xTrace4(havalp,TR_EVENTS,"havalOpen calling xOpen with args %x %x %x %x",
	      self,hlpType,xGetDown(self, TRANSPORT_PROTOCOL),part);

      lls = xOpen(self, hlpType, xGetDown(self, TRANSPORT_PROTOCOL), part);
      if ( lls == ERR_XOBJ )
	{ xTrace0(havalp, TR_ERRORS, "havalOpen: could not open lls"); return ERR_XOBJ; }
      if (self->numdown > 1) xCopyParts(&part2, &b, 0, lls);
    }

    if ( mapResolve(pstate->actMap, &lls, &havalSessn) == XK_SUCCESS ) {
        xTrace0(havalp, TR_MAJOR_EVENTS, "found an existing one");
        xClose(lls);
    } else {
        xTrace0(havalp, TR_MAJOR_EVENTS, "creating a new one");
        if ( ((havalSessn = havalCreateSessn(self, hlpRcv, hlpType, lls, part2)) == ERR_XOBJ))
	  {
	    xTrace0(havalp, TR_ERRORS, "havalOpen: create session failed");
	    if (b) xFree(b);
	    xClose(lls);
	  }
      }
    return (havalSessn);
}

/*
 *  havalCreateSessn
 *
 *	participants must be a copy of those used in creating lls
 *
 */
static XObj
havalCreateSessn( self, hlpRcv, hlpType, lls, part)
    XObj	self, hlpRcv, hlpType, lls;
    Part	*part;
{
    XObj	s;
    SState	*ss;
    PState	*ps = (PState *)self->state;
    XObj        SessnArray[3];
    int		down_count = 1;

    xTrace4(havalp, TR_EVENTS, "havalCreateSessn entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, lls);

    ss = X_NEW(SState);
    xTrace1(havalp, TR_FULL_TRACE, "havalCreateSessn: X_NEW(SState) returned %x", ss);

    SessnArray[0]=lls;

    ss->flags = 0;
    ss->LprefixL = 0; ss->Lprefix = (unsigned char *)0;
    ss->RprefixL = 0; ss->Rprefix = (unsigned char *)0;
   
    if (!(ps->flags & USE_IPSEC))
    {

        if (self->numdown > 1) {

            if (!(ps->flags & ALL_PARTS)) {

		Part part1[1], part2[1];
		XObj kms1, kms2;

		xTrace0(havalp, TR_DETAILED, "haval createSessn two keym session");

		/* open two key manager sessions */
		partInit(part1,1);
		part1[0] = part[REMOTE_PART];
		part1[0].len =1; /* major hack */
		kms1 = xOpen(self, self, xGetDown(self,KM_PROTL), part1);
		SessnArray[KM_PROTL_IN] = kms1;
		if (kms1 == ERR_XOBJ) { 
		  xTrace0(havalp, TR_ERRORS, "haval CreateSessn: could not open key manager");
		  xFree((char *) ss);
		  return ERR_XOBJ;
	    	}

	    	xTrace0(havalp, TR_DETAILED, "haval createSessn keym initialized");

	   	partInit(part2,1);
	    	part2[0] = part[LOCAL_PART];
	    	part2[0].len =1; /* major hack */
	    	kms2 = xOpen(self, self, xGetDown(self,KM_PROTL), part2);
	    	SessnArray[KM_PROTL_OUT] = kms2;
	    	if (kms2 == ERR_XOBJ) { 
	    		xTrace0(havalp, TR_ERRORS, "haval CreateSessn: could not open key manager");
	  		xClose(kms1);
	  		xFree((char *) ss);
	  		return ERR_XOBJ;
	         } 

	         xTrace0(havalp, TR_DETAILED, "haval createSessn keym initialized");

	         down_count = 3;
                 }
             else {
	         xTrace0(havalp, TR_DETAILED, "haval createSessn one keym sessn");
	         SessnArray[KM_PROTL] = xOpen(self, self, xGetDown(self, KM_PROTL), part);
	         if (SessnArray[KM_PROTL] == ERR_XOBJ) {
	  	     xTrace0(havalp, TR_ERRORS, "havalCreateSessn keymgr open error");
	             } 
                 down_count = 2;
             }
        }
    }
    else
    {
	Part p[1];
	int  lsaid;
	IPhost myhost;

     	if (partStackTopByteLen(part[1]) != sizeof(IPhost))
     	{
 		xTrace0(havalp, TR_MAJOR_EVENTS,
			"MD5 open error: ipsecmode: local host is not specified or not an IPhost");
		return ERR_XOBJ;
     	}
     	myhost=*(IPhost *)partPop(part[1]);
     	partPush(part[1], (char *)&myhost, sizeof(IPhost));
	lsaid = *((XSaid *)partPop(part[2]));
 
	xTrace1(havalp, TR_DETAILED, "haval createSessn one keym session using lsaid %u",
		 lsaid);
   	partInit(p,1);
     	partPush(p[0], (char *)&lsaid,sizeof(XSaid));	
     	partPush(p[0], (char *)&myhost, sizeof(IPhost));
	SessnArray[1] = xOpen(self, self, xGetDown(self,KM_PROTL), p);
	if (SessnArray[1] == ERR_XOBJ) { 
	    xTrace0(havalp, TR_ERRORS, "haval CreateSessn: could not open key manager");
	    return ERR_XOBJ;
            }
        down_count = 2;
    }
    s = xCreateSessn(havalSessnInit, hlpRcv, hlpType, self, down_count,
		     SessnArray);
    s->state = (VOID *)ss;
    s->binding = mapBind(ps->actMap, &lls, s);

    if ( s->binding == ERR_BIND ) {
	xTrace0(havalp, TR_ERRORS, "havalCreateSessn: mapBind error");
	xDestroy(s);
	xFree((char *) ss);
	return ERR_XOBJ;
      }

    xTrace1(havalp, TR_MORE_EVENTS, "havalCreateSessn returning %x",s);
    return s;
}


static xkern_return_t
havalOpenEnable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(havalp, TR_EVENTS, "havalOpenEnable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    if ( defaultOpenEnable(ps->pasMap, hlpRcv, hlpType, 
	(VOID *)&hlpType) == XK_FAILURE ) {
        xTrace0(havalp, TR_MORE_EVENTS, "havalOpenEnable defaultOpenEnable failed ");
        return XK_FAILURE;
    }
    

    if (self->numdown) {
	Part *qart;

      /* must copy part before using it */
      if (ps->flags & ALL_PARTS) xSimpleCopyParts(&qart, part);
      val = xOpenEnable(self, hlpType, xGetDown(self,TRANSPORT_PROTOCOL), part);
      if ( val == XK_FAILURE ) {
        xTrace0(havalp, TR_MORE_EVENTS, "havalOpenEnable xOpenEnable failed ");
        return XK_FAILURE;
      }
      if (ps->flags & ALL_PARTS) {
	val = xOpenEnable(self, self, xGetDown(self,KM_PROTL), qart);
	xFree((char *)qart);
	if (val == XK_FAILURE) { 
	  /* xOpenDisable(self, hlpType, xGetDown(self, TRANSPORT_PROTOCOL), part); */
	  xTrace0(havalp, TR_ERRORS, "HAVALOpenEnable: could not open key manager");
	  return XK_FAILURE;
	}
      }
    }
    else val = XK_SUCCESS;

    xTrace1(havalp, TR_MORE_EVENTS, "havalOpenEnable returning %x", val);
    return val;
} 


static xkern_return_t
havalOpenDisable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(havalp, TR_EVENTS, "havalOpenDisable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    val = defaultOpenDisable(ps->pasMap, hlpRcv, hlpType, (void*)&hlpType);

    xTrace1(havalp, TR_MORE_EVENTS, "havalOpenDisable returning %x", val);
    return val;
}

static xkern_return_t
havalOpenDone(self, lls, llp, hlpType)
    XObj self, lls, llp, hlpType;
{
    XObj        s;
    PState      *pstate;
    Enable      *e;
    Part	*p;
    char	*b = 0;
   
    xTrace0(havalp, TR_MAJOR_EVENTS, "Entered havalOpenDone");

    if ( self == hlpType ) {
        xTrace0(havalp, TR_ERRORS, "self == hlpType in havalOpenDone");
        return XK_FAILURE;
    }
    pstate = (PState *)self->state;

    /*
     * check for open enables
     */
    if ( mapResolve(pstate->pasMap, &hlpType, &e) == XK_FAILURE ) {
        /*
         * This shouldn't happen
         */
        xTrace0(havalp, TR_ERRORS,
                "havalOpenDone: Couldn't find hlp for incoming session");
        return XK_FAILURE;
    }

    xDuplicate(lls);
    xCopyParts(&p, &b, 0, lls);
    if ( (s = havalCreateSessn(self, e->hlpRcv, e->hlpType, lls, p))
                == ERR_XOBJ ) {
        xFree(b);
        xClose(lls);
        return XK_FAILURE;
    }

    xFree(b);
    xTrace0(havalp, TR_EVENTS,
            "havalOpenDone: Passively opened session successfully created");
    return xOpenDone(e->hlpRcv, s, self);
}

static xkern_return_t
havalDemux( self, lls, msg )
    XObj	self, lls;
    Msg		*msg;
{
  PState	*ps = (PState *)self->state;
  XObj 	s = 0;
  xkern_return_t val=XK_FAILURE;
  
  xTrace3(havalp, TR_EVENTS, "havalDemux entered: args %x %x %x",
	  self, lls, msg);

  xIfTrace(havalp, TR_DETAILED) msgShow(msg);

  if ( mapResolve(ps->actMap, &lls, &s) == XK_FAILURE ) { 
	xTrace0(havalp, TR_EVENTS, "havalDemux: map resolve fails");
  	return XK_FAILURE;
  }

  if ( xIsSession(s) )  val = xPop(s, lls, msg, 0);

  xIfTrace(havalp, TR_FUNCTIONAL_TRACE)
  {
  if (val == XK_SUCCESS)
  xTrace0(havalp, TR_FUNCTIONAL_TRACE, "havalDemux returning XK_SUCCESS");
  else if (val == XK_FAILURE)
  xTrace0(havalp, TR_FUNCTIONAL_TRACE, "havalDemux returning XK_FAILURE");
  else
  xTrace1(havalp, TR_FUNCTIONAL_TRACE, "havalDemux returning %x", val);
  }
  return val;
}


static xkern_return_t
havalClose( s )
    XObj	s;
{
    PState	*ps = (PState *)xMyProtl(s)->state;
    SState	*ss;
    xkern_return_t res;

    xTrace1(havalp, TR_MAJOR_EVENTS, "havalClose: closing session %x", s);
    xAssert( xIsSession( s ) );

    ss = (SState *)(s->state);
    if (ss->Lprefix && ss->Lprefix != ss->Rprefix)
      /* erase local prefix, in case it's sensitive */
    { if (ss->LprefixL) bzero((char *)ss->Lprefix,ss->LprefixL);
      xFree((VOID *)ss->Lprefix); }
    ss->LprefixL = 0; ss->Lprefix = 0;
    if (ss->Rprefix)  /* erase remote prefix, in case it's sensitive */
    { if (ss->RprefixL) bzero((char *)ss->Rprefix,ss->RprefixL);
      xFree((VOID *)ss->Rprefix); }
    ss->RprefixL = 0; ss->Rprefix = 0;
    if (s->numdown > 1) xClose(xGetDown(s, KM_PROTL));
    if (s->numdown > 2) xClose(xGetDown(s, KM_PROTL_IN));
    xAssert( s->rcnt == 0 );
    res = mapRemoveBinding( ps->actMap, s->binding );
    xAssert( res != XK_FAILURE );
    xClose(xGetDown(s,TRANSPORT_PROTOCOL));
    xDestroy( s );
    xTrace0(havalp, TR_DETAILED, "havalClose returning");
    return XK_SUCCESS;
}

static
long nullop(to,from,len,ignore)
    VOID *to; char *from; VOID *ignore; long len;
{
    xTrace4(havalp, TR_FULL_TRACE, "haval nullop args: %x %x %x %x",
	    to,from,len,ignore);

    return len;
}

static
long havalcopy(to,from,len,ignore)
    VOID *to; char *from; VOID *ignore; long len;
{
    xTrace4(havalp, TR_FULL_TRACE, "havalcopy args: %x %x %x %x",
	    to,from,len,ignore);

    bcopy(from,to,len);
    return len;
}

static
void havalStore(src, dst, len, arg)
    void *src;
    char *dst;
    long len;
    void *arg;
{
    xTrace4(havalp, TR_FULL_TRACE, "havalStore args: %x %x %x %x",
	    src,dst,len,arg);

    bcopy((char *)src,dst,len);
}


static int
havalchunk( msgptr, msglen, havalctxp )
  char *msgptr; long msglen; HAVAL_CTX *havalctxp;
{ 
  HAVALUpdate(havalctxp,(unsigned char *)msgptr,(unsigned int)msglen);
  return TRUE;
 }


static xmsg_handle_t
havalPush( s, msg )
    XObj	s;
    Msg 	*msg;
{
    long i;
    HAVAL_CTX havalctx;
    u_char digest[FPTLEN>>3];
    SState *ss;

    xTrace2(havalp, TR_EVENTS, "havalPush: session %x  msg %x", s, msg);
    xAssert(xIsSession(s));
    ss = (SState *)(s->state);

    HAVALInit(&havalctx);
    /* a prefix is used if we have a key manager or if the control op
       has been issued */
    if (ss->LprefixL && ss->Lprefix) 
    {
       HAVALUpdate(&havalctx,ss->Lprefix,ss->LprefixL);
       msgForEach(msg,(XCharFun)havalchunk,&havalctx); /* Feed each msg segment to HAVALUpdate. */
       if (((PState *)(s->myprotl->state))->flags & HAVAL_SECRETAPPEND)
       	   HAVALUpdate(&havalctx,ss->Lprefix,ss->LprefixL);
    }
    else  if (s->numdown > 1) {
	/* need to branch on ALL_PARTS */
	    int  LprefixL;
	    char Lprefix[MAX_PREFIX];
	    bzero(Lprefix, MAX_PREFIX);
	    LprefixL = xControl(xGetDown(s, KM_PROTL), KM_RESOLVE, Lprefix, MAX_PREFIX);
	    if (LprefixL > 0) HAVALUpdate(&havalctx,(unsigned char *)Lprefix,LprefixL);
     	    msgForEach(msg,(XCharFun)havalchunk,&havalctx); /* Feed each msg segment to HAVALUpdate. */
 	    if ((((PState *)(s->myprotl->state))->flags & HAVAL_SECRETAPPEND)&&(LprefixL > 0))
		 HAVALUpdate(&havalctx,(unsigned char *)Lprefix,LprefixL);
    }
    else
        msgForEach(msg,(XCharFun)havalchunk,&havalctx); /* Feed each msg segment to HAVALUpdate. */
    HAVALFinal(&havalctx, digest);
    if (!(((PState *)(s->myprotl->state))->flags & HAVAL_TRAILERS))
      msgPush(msg,havalStore,digest,FPTLEN>>3,0);  /* Push digest onto msg. */
    else
      msgAppend(msg,havalStore,digest,FPTLEN>>3,0,50); /* check the last arg;
							  want to leave enough
							  for ipsp hdr, ip hdr
							  */
    xIfTrace (havalp, TR_MORE_EVENTS)
    {  printf("havalPush: message length %d; digest is ", msgLen(msg));
       for (i=0;i<(FPTLEN>>3);i++) printf(" %02x", digest[i]);
       printf("\n"); }

    if (s->numdown)
      return xPush(xGetDown(s, TRANSPORT_PROTOCOL), msg);
    return XMSG_NULL_HANDLE;
}



static xkern_return_t
havalPop( s, lls, msg, hdr)
    XObj	s, lls;
    Msg 	*msg;
    void        *hdr;
{
    long len; int i;
    HAVAL_CTX havalctx;
    u_char digest[FPTLEN>>3];
    u_char calc_digest[FPTLEN>>3];
    SState *ss;

    xTrace4(havalp, TR_EVENTS, "havalPop: args %x %x %x %x", s, lls, msg, hdr);
    ss = (SState *)(s->state);

    len = msgLen(msg);
    if (len<FPTLEN>>3)
    { xTrace1(havalp, TR_ERRORS,
	      "havalPop received bad message, length %D ",len);
      return XK_FAILURE; }

    if (!(((PState *)(s->myprotl->state))->flags & HAVAL_TRAILERS))
      msgPop(msg, havalcopy, digest, FPTLEN>>3, 0);  /* read message digest */
    else {
      Msg tailmsg;
      long length;
      
      msgConstructCopy(&tailmsg, msg);
      msgPop(&tailmsg, nullop, digest, (length =  msgLen(msg)-(FPTLEN>>3)), 0);
      msgPop(&tailmsg, havalcopy, digest, FPTLEN>>3, 0);
      msgAssign(&tailmsg, msg);
      msgChopOff(&tailmsg, msg, length);
      msgDestroy(&tailmsg);
    }

    HAVALInit(&havalctx);
    if (ss->RprefixL && ss->Rprefix) 
    {
	HAVALUpdate(&havalctx,ss->Rprefix,ss->RprefixL);
       	msgForEach(msg,(XCharFun)havalchunk,&havalctx); /* Feed each msg segment to HAVALUpdate. */
        if (((PState *)(s->myprotl->state))->flags & HAVAL_SECRETAPPEND)
		HAVALUpdate(&havalctx,ss->Rprefix,ss->RprefixL);
    }
    else if (s->numdown > 1) {
	    int  RprefixL;
	    char Rprefix[MAX_PREFIX];
	    
	    bzero(Rprefix, MAX_PREFIX);
	    if (s->numdown > 2)
	      RprefixL = xControl(xGetDown(s,KM_PROTL_IN), KM_RESOLVE,
				  Rprefix, MAX_PREFIX);
	    else
	      RprefixL = xControl(xGetDown(s,KM_PROTL), KM_RESOLVE,
				  Rprefix, MAX_PREFIX);
	    if (RprefixL > 0)
	      HAVALUpdate(&havalctx,(unsigned char *)Rprefix,RprefixL);
       	    msgForEach(msg,(XCharFun)havalchunk,&havalctx); /* Feed each msg segment to HAVALUpdate. */
	    if ((((PState *)(s->myprotl->state))->flags & HAVAL_SECRETAPPEND)&&(RprefixL > 0))
	      HAVALUpdate(&havalctx,(unsigned char *)Rprefix,RprefixL);
    }
    else
       	msgForEach(msg,(XCharFun)havalchunk,&havalctx); /* Feed each msg segment to HAVALUpdate. */
    HAVALFinal(&havalctx, calc_digest);

    xIfTrace (havalp, TR_MORE_EVENTS)
    {  printf("havalPop: message length %d; digest is ", len);
       for (i=0;i<(FPTLEN>>3);i++) printf(" %02x", digest[i]);
       printf("\n"); }

    for (i=0;i<(FPTLEN>>3);i++)  /* compare received & computed digests */
    {  if ( digest[i] != calc_digest[i] )
       {  xTrace0(havalp, TR_ERRORS, "havalPop: bad compare for message digest");
	  xIfTrace (havalp, TR_EVENTS)
	  {  printf("havalPop: digest from header is ");
	     for (i=0;i<(FPTLEN>>3);i++) printf(" %02x", digest[i]);
	     printf("\nhavalPop: computed digest is    ");
	     for (i=0;i<(FPTLEN>>3);i++) printf(" %02x", calc_digest[i]);
	     printf("\n"); }
	  xTrace0(havalp, TR_ERRORS, "havalPop: invalid signature - dropping message");
	  return XK_FAILURE;
	}
     }

    /* Fixup length in IP pseudoheader? */
    if (((SState *)(s->state))->flags & IP_PSHDR_FLAG
	|| ((PState *)(xMyProtl(s)->state))->flags & IP_PSHDR_FLAG)
    {  IPpseudoHdr *ipphdr = msgGetAttr( msg, 0);
       if (ipphdr)
       {  ipphdr->len = htons(ntohs(ipphdr->len)-(FPTLEN>>3));
	  xTrace3(havalp, TR_FULL_TRACE,
	    "havalPop:  Adjusting IP pseudoheader length; ipphdr %x  phlen after %d  len before %d",
		  ipphdr, ntohs(ipphdr->len), len);  }
       else { xTrace0(havalp, TR_FULL_TRACE,
	       "havalPop:  Skipping length adjustment for IP pseudoheader - msg attr is 0."); };
     }

    return xDemux(s, msg);    
} 


static int
havalControlSessn(s, op, buf, len)
    XObj	s;
    int 	op, len;
    char 	*buf;
{
    xAssert(xIsSession(s));
    switch (op) {

      case GETMAXPACKET:
      case GETOPTPACKET:
	{ 
	    int max;

	    if (!s->numdown) return -1;
	    if (len < sizeof(max) || !buf) return 0; 
	    if (xControl(xGetDown(s, 0), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
	    max -= (FPTLEN>>3); if (max<0) return 0;
	    *(int *)buf = max;
	    return(sizeof(max));
	}

      case HASH_SIZE:
	return HAVAL_SIZE;

      case HASH_SETPREFIX:
	{  SState *ss = (SState *)(s->state);

	   if (ss->Lprefix && ss->Lprefix != ss->Rprefix)
	     /* erase old prefix, in case it's sensitive */
	     { if (ss->LprefixL) bzero((char *)ss->Lprefix,ss->LprefixL); xFree((VOID *)ss->Lprefix); }
	   if (ss->Rprefix)  /* erase old prefix, in case it's sensitive */
	     { if (ss->RprefixL) bzero((char *)ss->Rprefix,ss->RprefixL); xFree((VOID *)ss->Rprefix); }
	   if (len > 0)
	   {  ss->LprefixL = ss->RprefixL = len;
	      ss->Lprefix =  ss->Rprefix = (unsigned char *)xMalloc(len);
	      bcopy(buf,(char *)ss->Lprefix,len); }
	   else { ss->LprefixL =  ss->RprefixL = 0; ss->Lprefix = ss->Rprefix = 0; }
	   return 0;  }

      case HASH_SETLOCALPREFIX:
	{  SState *ss = (SState *)(s->state);

	   if (ss->Lprefix && ss->Lprefix != ss->Rprefix)
	     /* erase old prefix, in case it's sensitive */
	     { if (ss->LprefixL) bzero((char *)ss->Lprefix,ss->LprefixL); xFree((VOID *)ss->Lprefix); }
	   if (len > 0)
	   {  ss->LprefixL = len; ss->Lprefix = (unsigned char *)xMalloc(len);
	      bcopy(buf,(char *)ss->Lprefix,len);  }
	   else { ss->LprefixL = 0; ss->Lprefix = 0; }
	   return 0;  }

      case HASH_SETREMOTEPREFIX:
	{  SState *ss = (SState *)(s->state);

	   if (ss->Rprefix && ss->Rprefix != ss->Lprefix)
	     /* erase old prefix, in case it's sensitive */
	     { if (ss->RprefixL) bzero((char *)ss->Rprefix,ss->RprefixL); xFree((VOID *)ss->Rprefix); }
	   if (len > 0)
	   {  ss->RprefixL = len; ss->Rprefix = (unsigned char *)xMalloc(len);
	      bcopy(buf,(char *)ss->Rprefix,len);  }
	   else { ss->RprefixL = 0; ss->Rprefix = 0; }
	   return 0;  }

      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((SState *)(s->state))->flags |= IP_PSHDR_FLAG;
	    xTrace0(havalp, TR_EVENTS,
	       "havalControlSessn: Turning on IPpseudoheader length fixup.");  }
	    if (!s->numdown) return 0;
      /* **** FALL THROUGH TO DEFAULT CASE, so lower sessions see it too. **** */

      default:
        if (buf && len) bzero(buf, len);   /* prevent info going down */
	if (s->numdown)
	  return xControl(xGetDown(s,TRANSPORT_PROTOCOL), op, buf, len);
	else return -1;
    }
}


static int
havalControlProtl( self, op, buf, len )
    XObj	self;
    int 	op, len;
    char 	*buf;
{
    xAssert(xIsProtocol(self));

    switch (op) {
      case HASH_SIZE:
	return HAVAL_SIZE;

      case GETMAXPACKET:
      case GETOPTPACKET:
        {
            int max;
	    if (len < sizeof(max) || !buf) return 0; 
	    if (xControl(xGetDown(self, TRANSPORT_PROTOCOL), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
	    max -= (FPTLEN>>3) ; if (max<0) return 0;
            *(int *)buf = max;
            return(sizeof(max));
	}

      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((PState *)(self->state))->flags |= IP_PSHDR_FLAG;
	    xTrace0(havalp, TR_EVENTS,
	       "havalControlProtl: Turning on IPpseudoheader length fixup.");  }
      /* **** FALL THROUGH TO DEFAULT CASE, so lower protocols see it too. **** */

      default:
	if (self->numdown) {
	  if (buf && len) bzero(buf, len);   /* prevent info going down */
	  return xControl(xGetDown(self, TRANSPORT_PROTOCOL), op, buf, len);
	}
	else return -1;
  }
  
}


static void
havalSessnInit(s)
    XObj s;
{
  s->push = havalPush;
  s->pop = havalPop;
  s->close = havalClose;
  s->control = havalControlSessn;
}


