/*
 * $RCSfile: hmac.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 *
 *
 */


/* The HMAC protocol pushes a 16byte cryptographic checksum onto the front
   of a message.  On the incoming side, it pops a 16byte checksum, and
   compares it to the checksum computed for the rest of the message.
   The code for HMAC is directly from the distribution.
*/


#include "xkernel.h"
#include "ip.h"                /* For IP pseudoheader fixup. */
#include "hmac_i.h"
#include "md5hash.h"
#include "km.h"
#include "hash.h"
#include "saidmgr.h"

#include "hmac_hash.c"

extern int tracehmacp; int tracehmacp;

/*  Byte order note:  All data is grouped in the order ABCD for HMAC processing. */


#ifdef __STDC__

static XObj		hmacCreateSessn( XObj, XObj, XObj, XObj, Part *);
static void		hmacSessnInit( XObj );
static int		hmacControlProtl( XObj, int, char *, int );
static xkern_return_t	hmacDemux( XObj, XObj, Msg * );
static XObj		hmacOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	hmacOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	hmacOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	hmacOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t	hmacClose( XObj );
static xmsg_handle_t	hmacPush( XObj, Msg * );
static xkern_return_t	hmacPop( XObj, XObj, Msg *, void * );
static void 		hmacStore( void *, char *, long, void * );
static int		hmacControlSessn( XObj, int, char *, int );
static int		hmacchunk( char *, long, MD5_CTX * );
static long 		hmaccopy(void *, char*, long, void * );
static long 		nullop(void *, char*, long, void * );
int			xk_hmac(unsigned char [], unsigned char *, int, int, unsigned char []);
void hmac_init( XObj );
static int		xCopyParts(Part **, char **, Part *, XObj);
static void		xSimpleCopyParts(Part **, Part *);

#else

static XObj		hmacCreateSessn();
static void		hmacSessnInit();
static int		hmacControlProtl();
static xkern_return_t	hmacDemux();
static XObj		hmacOpen();
static xkern_return_t	hmacOpenEnable();
static xkern_return_t	hmacOpenDisable();
static xkern_return_t	hmacOpenDone();
static xkern_return_t	hmacClose();
static xmsg_handle_t	hmacPush();
static xkern_return_t	hmacPop();
static long 		hmacStore();
static long 		hmaccopy();
static long 		nullop();
static int		hmacControlSessn();
int			xk_hmac();
static int		xCopyParts();
static void		xSimpleCopyParts();

#endif /* __STDC__ */

static XObj errobj;

static xkern_return_t romfindpartnum(XObj, char **, int, int, void *);
static xkern_return_t romuseipsec(XObj, char **, int, int, void *);
static xkern_return_t romleafmode(XObj, char **, int, int, void *);
static xkern_return_t romfindsecretappend(XObj, char **, int, int, void *);
static xkern_return_t romfindtrailer(XObj, char **, int, int, void *);
static xkern_return_t romrfc1828(XObj, char **, int, int, void *);

static XObjRomOpt rmopts[6] = {{ "numparts", 1, romfindpartnum },
			       { "leafmode", 1, romleafmode },
			       { "useipsec", 1, romuseipsec },
			       { "secretappend", 1, romfindsecretappend },
			       { "trailers", 1, romfindtrailer },
			       { "rfc1828", 1, romrfc1828 }};

static xkern_return_t
romfindpartnum(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  if (atoi(ptr[2]) > 1)
    ((PState *)(self->state))->flags |= ALL_PARTS;
  else
    ((PState *)(self->state))->flags &= ~ALL_PARTS;
  return XK_SUCCESS;
}

static xkern_return_t
romuseipsec(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
    ((PState *)(self->state))->flags |= USE_IPSEC;
    return XK_SUCCESS;
}

static xkern_return_t
romleafmode(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
    ((PState *)(self->state))->noleafmode = 0;
    return XK_SUCCESS;
}

static xkern_return_t
romfindsecretappend(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  ((PState *)(self->state))->flags |= MD5_SECRETAPPEND;
  return XK_SUCCESS;
}

static xkern_return_t
romfindtrailer(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  xTrace0(hmacp, TR_GROSS_EVENTS, "hmac rom opt turning on trailer");
  ((PState *)(self->state))->flags |= MD5_TRAILERS;
  return XK_SUCCESS;
}

static xkern_return_t
romrfc1828(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  xTrace0(hmacp, TR_GROSS_EVENTS, "hmac rom opt turning on rfc1828");
  ((PState *)(self->state))->flags |= MD5_SECRETAPPEND;
  ((PState *)(self->state))->flags |= RFC1828;
  return XK_SUCCESS;
}

void
hmac_init( self )
     XObj self;
{
    PState	*ps;
    XObj	llp;

    xTrace0(hmacp, TR_ERRORS, "hmac_init entered");
    xAssert(xIsProtocol(self));
    
    if ( self->numdown &&
	! xIsProtocol(llp = xGetDown(self, TRANSPORT_PROTOCOL)) )
      { xError("hmac_init: hmac cannot get lower protocol object");
	/* return; */}
    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->actMap = mapCreate(MD5_ACTIVE_MAP_SZ, sizeof(ActiveId));
    ps->pasMap = mapCreate(MD5_PASSIVE_MAP_SZ, sizeof(PassiveId));
    ps->flags = 0;
    ps->noleafmode = 1;

    self->control = hmacControlProtl;
    self->open = hmacOpen;
    self->openenable = hmacOpenEnable;
    self->opendisable = hmacOpenDisable;
    self->opendone = hmacOpenDone;
    self->close = hmacClose;  
    self->demux = hmacDemux;
    bzero((char *)&errobj, sizeof(XObj));
    findXObjRomOpts(self, rmopts, sizeof(rmopts)/sizeof(rmopts[0]), 0);
    if ((!(self->numdown)) && (ps->noleafmode))  
  	Kabort("hmac_init: no transport protocol in non leaf mode");
    xTrace0(hmacp, TR_EVENTS, "hmac_init exiting");
}

static int
xCopyParts(q, qbuf, p, lls)
     Part **q;
     char **qbuf;
     Part *p;
     XObj lls;
{
  char *ebuf;
  int	len = 500;

  ebuf = xMalloc(500);
  if (lls != ERR_XOBJ) {
    if (xControl(lls, GETPARTICIPANTS, ebuf, len)<0)
      return 0;
  }
  else
    partExternalize(p, ebuf, &len);
  *q = (Part *)xMalloc(partExtLen(ebuf)*sizeof(Part));
  partInternalize(*q, ebuf);
  *qbuf = ebuf;
  return partLen(((Part *)(*q)));
}

static void
xSimpleCopyParts(q, p)
     Part **q;
     Part *p;
{
  int  pl =  partLen(p);

  *q = (Part *)xMalloc(pl*sizeof(Part));
  partInit(*q, pl);
  while (pl--) { (*q)[pl] = p[pl]; }
}

static XObj
hmacOpen( self, hlpRcv, hlpType, part )
    XObj	self, hlpRcv, hlpType;
    Part 	*part;
{
    XObj 	hmacSessn;
    XObj	lls = errobj;
    PState	*pstate;
    Part	*part2 = 0;
    char	*b = 0;

    pstate = (PState *) self->state;


    if (pstate->noleafmode) {
      xTrace4(hmacp,TR_EVENTS,"hmacOpen calling xOpen with args %x %x %x %x",
	      self,hlpType,xGetDown(self, TRANSPORT_PROTOCOL),part);

      lls = xOpen(self, hlpType, xGetDown(self, TRANSPORT_PROTOCOL), part);
      if ( lls == ERR_XOBJ )
	{ xTrace0(hmacp, TR_ERRORS, "hmacOpen: could not open lls"); return ERR_XOBJ; }
      if (self->numdown > 1) xCopyParts(&part2, &b, 0, lls);
    }

    if ( mapResolve(pstate->actMap, &lls, &hmacSessn) == XK_SUCCESS ) {
      xTrace0(hmacp, TR_MAJOR_EVENTS, "hmacp: found an existing one");
      if (lls != errobj) xClose(lls);
    } else {
      xTrace0(hmacp, TR_MAJOR_EVENTS, "hmacp: creating a new one");
      if (!part2) xSimpleCopyParts(&part2, part);
      if ( (hmacSessn = hmacCreateSessn(self, hlpRcv, hlpType, lls, part2))
	  == ERR_XOBJ )
	{
	  xTrace0(hmacp, TR_ERRORS, "hmacOpen: create session failed");
	  if (b) xFree(b);
	  xClose(lls);
	}
    }
    return (hmacSessn);
}

/*
 *  hmacCreateSessn
 *
 *	participants must be a copy of those used in creating lls
 *
 */
static XObj
hmacCreateSessn( self, hlpRcv, hlpType, lls, part)
    XObj	self, hlpRcv, hlpType, lls;
    Part	*part;
{
    XObj	s;
    SState	*ss;
    PState	*ps = (PState *)self->state;
    XObj        SessnArray[3];
    int		down_count;

    xTrace4(hmacp, TR_EVENTS, "hmacCreateSessn entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, lls);

    ss = X_NEW(SState);
    xTrace1(hmacp, TR_FULL_TRACE, "hmacCreateSessn: X_NEW(SState) returned %x", ss);

    if (ps->noleafmode)
    {
    	SessnArray[0]=lls;
	down_count = 1;
    } 
    else 
    {
    	SessnArray[0]=NULL;
	down_count = 0;
    }

    ss->flags = 0;
    ss->LprefixL = 0; ss->Lprefix = (unsigned char *)0;
    ss->RprefixL = 0; ss->Rprefix = (unsigned char *)0;
    ss->firstblock = (unsigned char *)0;

    if (self->numdown > 0 && !ps->noleafmode) {
        if (!(ps->flags & USE_IPSEC))
        {
 
      	    if (!(ps->flags & ALL_PARTS)) {

		Part part1[1], part2[1];
		XObj kms1, kms2;

		xTrace0(hmacp, TR_DETAILED, "hmac createSessn two keym session");

		/* open two key manager sessions */
		partInit(part1,1);
		part1[0] = part[REMOTE_PART];
		part1[0].len =1; /* major hack */
		kms1 = xOpen(self, self, xGetDown(self,ps->noleafmode), part1);
		SessnArray[KM_PROTL_IN] = kms1;
		if (kms1 == ERR_XOBJ) { 
		  xTrace0(hmacp, TR_ERRORS, "hmac CreateSessn: could not open key manager");
		  xFree((char *) ss);
		  return ERR_XOBJ;
		}

		xTrace0(hmacp, TR_DETAILED, "hmac createSessn keym initialized");

		partInit(part2,1);
		part2[0] = part[LOCAL_PART];
		part2[0].len =1; /* major hack */
		kms2 = xOpen(self, self, xGetDown(self,ps->noleafmode), part2);
		SessnArray[KM_PROTL_OUT] = kms2;
		if (kms2 == ERR_XOBJ) { 
		  xTrace0(hmacp, TR_ERRORS, "hmac CreateSessn: could not open key manager");
		  xClose(kms1);
		  xFree((char *) ss);
		  return ERR_XOBJ;
		} 

		xTrace0(hmacp, TR_DETAILED, "hmac createSessn keym initialized");

		down_count = 3;
          }
          else {
	         xTrace0(hmacp, TR_DETAILED, "hmac createSessn one keym sessn");
	         SessnArray[KM_PROTL] = xOpen(self, self, xGetDown(self, ps->noleafmode), part);
	         if (SessnArray[KM_PROTL] == ERR_XOBJ) {
	  	     xTrace0(hmacp, TR_ERRORS, "hmacCreateSessn keymgr open error");
	             } 
                 down_count = 2;
	  }
       }
    	else
    	{
	 	Part p[1];
		XSaid  lsaid;
		IPhost spihost;

#ifdef 0
     		if (partStackTopByteLen(part[1]) != sizeof(IPhost))
     		{
 			xTrace0(hmacp, TR_MAJOR_EVENTS,
		    	"HMAC open error: ipsecmode: local host is not specified or not an IPhost");
			return ERR_XOBJ;
     		}

     		myhost=*(IPhost *)partPop(part[1]);
     		partPush(part[1], (char *)&myhost, sizeof(IPhost));
#endif /* 0 */
     		spihost=*(IPhost *)partPop(part[SPI_PART]);
		lsaid = *((XSaid *)partPop(part[SPI_PART]));
		
		xTrace0(hmacp, TR_DETAILED, "hmac createSessn one keym sessn using lsaid");
     		partInit(p,1);
     		partPush(p[0], (char *)&lsaid,sizeof(XSaid));	
     		partPush(p[0], (char *)&spihost, sizeof(IPhost));
		SessnArray[1] = xOpen(self, self, xGetDown(self, ps->noleafmode), p);
		if (SessnArray[1] == ERR_XOBJ) {
		 	xTrace0(hmacp, TR_ERRORS, "hmacCreateSessn keymgr open error");
	        	return ERR_XOBJ;
		}
		down_count = 2;
    	}
    }

    if (ps->flags & RFC1828)
      ss->firstblock = xMalloc(64);
    s = xCreateSessn(hmacSessnInit, hlpRcv, hlpType, self, down_count,
		     SessnArray);
    s->state = (VOID *)ss;
    s->binding = mapBind(ps->actMap, &lls, s);

    if ( s->binding == ERR_BIND ) {
	xTrace0(hmacp, TR_ERRORS, "hmacCreateSessn: mapBind error");
	xDestroy(s);
	if (ss->firstblock) xFree(ss->firstblock);
	xFree((char *) ss);
	return ERR_XOBJ;
      }

    xTrace1(hmacp, TR_MORE_EVENTS, "hmacCreateSessn returning %x",s);
    return s;
}


static xkern_return_t
hmacOpenEnable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(hmacp, TR_EVENTS, "hmacOpenEnable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    if ( defaultOpenEnable(ps->pasMap, hlpRcv, hlpType, 
	(VOID *)&hlpType) == XK_FAILURE ) {
        xTrace0(hmacp, TR_MORE_EVENTS, "hmacOpenEnable defaultOpenEnable failed ");
        return XK_FAILURE;
    }
    

    if (self->numdown) {
	Part *qart;

      /* must copy part before using it */
      if (ps->flags & ALL_PARTS) xSimpleCopyParts(&qart, part);
      val = xOpenEnable(self, hlpType, xGetDown(self,TRANSPORT_PROTOCOL), part);
      if ( val == XK_FAILURE ) {
        xTrace0(hmacp, TR_MORE_EVENTS, "hmacOpenEnable xOpenEnable failed ");
        return XK_FAILURE;
      }
      if (ps->flags & ALL_PARTS) {
	val = xOpenEnable(self, self, xGetDown(self,KM_PROTL), qart);
	xFree((char *)qart);
	if (val == XK_FAILURE) { 
	  /* xOpenDisable(self, hlpType, xGetDown(self, TRANSPORT_PROTOCOL), part); */
	  xTrace0(hmacp, TR_ERRORS, "HMACOpenEnable: could not open key manager");
	  return XK_FAILURE;
	}
      }
    }
    else val = XK_SUCCESS;

    xTrace1(hmacp, TR_MORE_EVENTS, "hmacOpenEnable returning %x", val);
    return val;
} 


static xkern_return_t
hmacOpenDisable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(hmacp, TR_EVENTS, "hmacOpenDisable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    val = defaultOpenDisable(ps->pasMap, hlpRcv, hlpType, (void*)&hlpType);

    xTrace1(hmacp, TR_MORE_EVENTS, "hmacOpenDisable returning %x", val);
    return val;
}

static xkern_return_t
hmacOpenDone(self, lls, llp, hlpType)
    XObj self, lls, llp, hlpType;
{
    XObj        s;
    PState      *pstate;
    Enable      *e;
    Part	*p;
    char	*b = 0;
   
    xTrace0(hmacp, TR_MAJOR_EVENTS, "Entered hmacOpenDone");

    if ( self == hlpType ) {
        xTrace0(hmacp, TR_ERRORS, "self == hlpType in hmacOpenDone");
        return XK_FAILURE;
    }
    pstate = (PState *)self->state;

    /*
     * check for open enables
     */
    if ( mapResolve(pstate->pasMap, &hlpType, &e) == XK_FAILURE ) {
        /*
         * This shouldn't happen
         */
        xTrace0(hmacp, TR_ERRORS,
                "hmacOpenDone: Couldn't find hlp for incoming session");
        return XK_FAILURE;
    }

    xDuplicate(lls);
    xCopyParts(&p, &b, 0, lls);
    if ( (s = hmacCreateSessn(self, e->hlpRcv, e->hlpType, lls, p))
                == ERR_XOBJ ) {
        xFree(b);
        xClose(lls);
        return XK_FAILURE;
    }

    xFree(b);
    xTrace0(hmacp, TR_EVENTS,
            "hmacOpenDone: Passively opened session successfully created");
    return xOpenDone(e->hlpRcv, s, self);
}

static xkern_return_t
hmacDemux( self, lls, msg )
    XObj	self, lls;
    Msg		*msg;
{
  PState	*ps = (PState *)self->state;
  XObj 	s = 0;
  xkern_return_t val = XK_FAILURE;
  
  xTrace3(hmacp, TR_EVENTS, "hmacDemux entered: args %x %x %x",
	  self, lls, msg);

  xIfTrace(hmacp, TR_DETAILED) msgShow(msg);

  if ( mapResolve(ps->actMap, &lls, &s) == XK_FAILURE ) { 
	xTrace0(hmacp, TR_EVENTS, "hmacDemux: map resolve fails");
  	return XK_FAILURE;
  }

  if ( xIsSession(s) )  val = xPop(s, lls, msg, 0);
  
  xIfTrace(hmacp, TR_FUNCTIONAL_TRACE)
  {
  if (val == XK_SUCCESS)
  xTrace0(hmacp, TR_FUNCTIONAL_TRACE, "hmacDemux returning XK_SUCCESS");
  else if (val == XK_FAILURE)
  xTrace0(hmacp, TR_FUNCTIONAL_TRACE, "hmacDemux returning XK_FAILURE");
  else
  xTrace1(hmacp, TR_FUNCTIONAL_TRACE, "hmacDemux returning %x", val);
  }

  return val;
}


static xkern_return_t
hmacClose( s )
    XObj	s;
{
    PState	*ps = (PState *)xMyProtl(s)->state;
    SState	*ss;
    xkern_return_t res;

    xTrace1(hmacp, TR_MAJOR_EVENTS, "hmacClose: closing session %x", s);
    xAssert( xIsSession( s ) );
/*  xClose(x);
*/
    ss = (SState *)(s->state);
    if (ss->Lprefix && ss->Lprefix != ss->Rprefix)
      /* erase local prefix, in case it's sensitive */
    { if (ss->LprefixL) bzero((char *)ss->Lprefix,ss->LprefixL);
      xFree((VOID *)ss->Lprefix); }
    ss->LprefixL = 0; ss->Lprefix = 0;
    if (ss->Rprefix)  /* erase remote prefix, in case it's sensitive */
    { if (ss->RprefixL) bzero((char *)ss->Rprefix,ss->RprefixL);
      xFree((VOID *)ss->Rprefix); }
    ss->RprefixL = 0; ss->Rprefix = 0;
    if ((ps->flags & RFC1828) && ss->firstblock)
      { bzero(ss->firstblock,64); xFree(ss->firstblock); }
    if (s->numdown > 1) xClose(xGetDown(s, KM_PROTL));
    if (s->numdown > 2) xClose(xGetDown(s, KM_PROTL_IN));
    xAssert( s->rcnt == 0 );
    res = mapRemoveBinding( ps->actMap, s->binding );
    xAssert( res != XK_FAILURE );
    if (ps->noleafmode)
    	xClose(xGetDown(s,TRANSPORT_PROTOCOL));
    xDestroy( s );
    xTrace0(hmacp, TR_DETAILED, "hmacClose returning");
    return XK_SUCCESS;
}

static
long nullop(to,from,len,ignore)
    VOID *to; char *from; VOID *ignore; long len;
{
    xTrace4(hmacp, TR_FULL_TRACE, "hmac nullop args: %x %x %x %x",
	    to,from,len,ignore);

    return len;
}

static
long hmaccopy(to,from,len,ignore)
    VOID *to; char *from; VOID *ignore; long len;
{
    xTrace4(hmacp, TR_FULL_TRACE, "hmaccopy args: %x %x %x %x",
	    to,from,len,ignore);

    bcopy(from,to,len);
    return len;
}

static
void hmacStore(src, dst, len, arg)
    void *src;
    char *dst;
    long len;
    void *arg;
{
    xTrace4(hmacp, TR_FULL_TRACE, "hmacStore args: %x %x %x %x",
	    src,dst,len,arg);

    bcopy((char *)src,dst,len);
}


static int
hmacchunk( msgptr, msglen, hmacctxp )
  char *msgptr; long msglen; MD5_CTX *hmacctxp;
{ 
  MD5Update(hmacctxp,(unsigned char *)msgptr,(unsigned int)msglen);
  return TRUE;
 }


static xmsg_handle_t
hmacPush( s, msg )
    XObj	s;
    Msg 	*msg;
{
    long i;
    MD5_CTX hmacctx;
    SState *ss;
    PState *ps;

    xTrace2(hmacp, TR_EVENTS, "hmacPush: session %x  msg %x", s, msg);
    xAssert(xIsSession(s));
    ss = (SState *)(s->state);
    ps = (PState *)(s->myprotl->state);
    /* a prefix is used if we have a key manager or if the control op
       has been issued */
    if (ss->LprefixL && ss->Lprefix) {
      hmac_md5(msg, ss->Lprefix, ss->LprefixL, (caddr_t) &hmacctx);
    } else if (s->numdown > 1) {
      /* need to branch on ALL_PARTS */
      int  LprefixL;
      char Lprefix[MAX_PREFIX];
      bzero(Lprefix, MAX_PREFIX);
      LprefixL = xControl(xGetDown(s, KM_PROTL), KM_RESOLVE, Lprefix, MAX_PREFIX);
      if(LprefixL > 0) {
	hmac_md5(msg, Lprefix, LprefixL, (caddr_t) &hmacctx);
      } else {
	MD5Init(&hmacctx);
	msgForEach(msg,(XCharFun)hmacchunk,&hmacctx); /* Feed each msg segment to MD5Update. */
	MD5Final(&hmacctx);
      }
      bzero(Lprefix, MAX_PREFIX); LprefixL = 0;  /* useless; compiler will drop */
    } else {
      MD5Init(&hmacctx);
      msgForEach(msg,(XCharFun)hmacchunk,&hmacctx); /* Feed each msg segment to MD5Update. */
      MD5Final(&hmacctx);
    }

    if (!(((PState *)(s->myprotl->state))->flags & MD5_TRAILERS))
      msgPush(msg,hmacStore,&hmacctx.digest[0],16,0);  /* Push digest onto msg. */
    else
      msgAppend(msg,hmacStore,&hmacctx.digest[0],16,0,50); /* check the last arg;
							  want to leave enough
							  for ipsp hdr, ip hdr
							  */
    xIfTrace (hmacp, TR_MORE_EVENTS)
      {  printf("hmacPush: message length %d; digest is ", msgLen(msg));
      for (i=0;i<16;i++) printf(" %02x", hmacctx.digest[i]);
      printf("\n"); }
    
    if (ps->noleafmode)
      return xPush(xGetDown(s, TRANSPORT_PROTOCOL), msg);
    return XMSG_NULL_HANDLE;
}



static xkern_return_t
hmacPop( s, lls, msg, hdr)
    XObj	s, lls;
    Msg 	*msg;
    void        *hdr;
{
    long len; int i;
    MD5_CTX hmacctx;
    u_char digest[16];
    SState *ss;
    PState *ps;

    xTrace4(hmacp, TR_EVENTS, "hmacPop: args %x %x %x %x", s, lls, msg, hdr);
    ss = (SState *)(s->state);
    ps = (PState *)(s->myprotl->state);

    len = msgLen(msg);
    if (len<16)
    { xTrace1(hmacp, TR_ERRORS,
	      "hmacPop received bad message, length %D less than 16",len);
      return XK_FAILURE; }

    if (!(((PState *)(s->myprotl->state))->flags & MD5_TRAILERS))
      msgPop(msg, hmaccopy, digest, 16, 0);  /* read message digest */
    else {
      Msg tailmsg;
      long leng;
      
      msgConstructCopy(&tailmsg, msg);
      msgPop(&tailmsg, nullop, digest, (leng = msgLen(msg)-16), 0);
      msgPop(&tailmsg, hmaccopy, digest, 16, 0);
      msgAssign(&tailmsg, msg);
      msgChopOff(&tailmsg, msg, leng);
      msgDestroy(&tailmsg);
    }

    if (ss->RprefixL && ss->Rprefix) {
      hmac_md5(msg, ss->Rprefix, ss->RprefixL, (caddr_t) &hmacctx);
    } else if (s->numdown > 1) {
      int  RprefixL;
      char Rprefix[MAX_PREFIX];
      bzero(Rprefix, MAX_PREFIX);
      if (s->numdown > 2)
	RprefixL = xControl(xGetDown(s,KM_PROTL_IN), KM_RESOLVE,
			    Rprefix, MAX_PREFIX);
      else
	RprefixL = xControl(xGetDown(s,KM_PROTL), KM_RESOLVE,
			    Rprefix, MAX_PREFIX);
      if (RprefixL > 0) {
	hmac_md5(msg, Rprefix, RprefixL, (caddr_t) &hmacctx);
      } else {
	MD5Init(&hmacctx);
	msgForEach(msg,(XCharFun)hmacchunk,&hmacctx); /* Feed each msg segment to MD5Update. */
	MD5Final(&hmacctx);
      }
      bzero(Rprefix, MAX_PREFIX); RprefixL = 0;  /* useless; compiler will drop */
    } else {
      MD5Init(&hmacctx);
      msgForEach(msg,(XCharFun)hmacchunk,&hmacctx); /* Feed each msg segment to MD5Update. */
      MD5Final(&hmacctx);
    }

    xIfTrace (hmacp, TR_MORE_EVENTS)
    {  printf("hmacPop: message length 16+%d; digest is ", len-16);
       for (i=0;i<16;i++) printf(" %02x", hmacctx.digest[i]);
       printf("\n"); }

    for (i=0;i<16;i++)  /* compare received & computed digests */
    {  if ( digest[i] != hmacctx.digest[i] )
       {  xTrace0(hmacp, TR_ERRORS, "hmacPop: bad compare for message digest");
	  xIfTrace (hmacp, TR_EVENTS)
	  {  printf("hmacPop: digest from header is ");
	     for (i=0;i<16;i++) printf(" %02x", digest[i]);
	     printf("\nhmacPop: computed digest is    ");
	     for (i=0;i<16;i++) printf(" %02x", hmacctx.digest[i]);
	     printf("\n"); }
	  xTrace0(hmacp, TR_ERRORS, "hmacPop: invalid signature - dropping message");
	  return XK_FAILURE;
	}
     }

    /* Fixup length in IP pseudoheader? */
    if (((SState *)(s->state))->flags & IP_PSHDR_FLAG
	|| ((PState *)(xMyProtl(s)->state))->flags & IP_PSHDR_FLAG)
    {  IPpseudoHdr *ipphdr = msgGetAttr( msg, 0);
       if (ipphdr)
       {  ipphdr->len = htons(ntohs(ipphdr->len)-16);
	  xTrace3(hmacp, TR_FULL_TRACE,
	    "hmacPop:  Adjusting IP pseudoheader length; ipphdr %x  phlen after %d  len before %d",
		  ipphdr, ntohs(ipphdr->len), len);  }
       else { xTrace0(hmacp, TR_FULL_TRACE,
	       "hmacPop:  Skipping length adjustment for IP pseudoheader - msg attr is 0."); };
     }

    if (ps->noleafmode)
    	return xDemux(s, msg);    
    else
	return XK_SUCCESS;
} 


static int
hmacControlSessn(s, op, buf, len)
    XObj	s;
    int 	op, len;
    char 	*buf;
{
    xAssert(xIsSession(s));
    switch (op) {

      case GETMAXPACKET:
      case GETOPTPACKET:
	{ 
	    int max;

	    if (!s->numdown) return -1;
	    if (len < sizeof(max) || !buf) return 0; 
	    if (xControl(xGetDown(s, 0), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
	    max -= 16; if (max<0) return 0;
	    *(int *)buf = max;
	    return(sizeof(max));
	}

      case HASH_SIZE:
	return MD5_SIZE;

      case HASH_SETPREFIX:
	{  SState *ss = (SState *)(s->state);

	   if (ss->Lprefix && ss->Lprefix != ss->Rprefix)
	     /* erase old prefix, in case it's sensitive */
	     { if (ss->LprefixL) bzero((char *)ss->Lprefix,ss->LprefixL); xFree((VOID *)ss->Lprefix); }
	   if (ss->Rprefix)  /* erase old prefix, in case it's sensitive */
	     { if (ss->RprefixL) bzero((char *)ss->Rprefix,ss->RprefixL); xFree((VOID *)ss->Rprefix); }
	   if (len > 0)
	   {  ss->LprefixL = ss->RprefixL = len;
	      ss->Lprefix =  ss->Rprefix = (unsigned char *)xMalloc(len);
	      bcopy(buf,(char *)ss->Lprefix,len); }
	   else { ss->LprefixL =  ss->RprefixL = 0; ss->Lprefix = ss->Rprefix = 0; }
	   return 0;  }

      case HASH_SETLOCALPREFIX:
	{  SState *ss = (SState *)(s->state);

	   if (ss->Lprefix && ss->Lprefix != ss->Rprefix)
	     /* erase old prefix, in case it's sensitive */
	     { if (ss->LprefixL) bzero((char *)ss->Lprefix,ss->LprefixL); xFree((VOID *)ss->Lprefix); }
	   if (len > 0)
	   {  ss->LprefixL = len; ss->Lprefix = (unsigned char *)xMalloc(len);
	      bcopy(buf,(char *)ss->Lprefix,len);  }
	   else { ss->LprefixL = 0; ss->Lprefix = 0; }
	   return 0;  }

      case HASH_SETREMOTEPREFIX:
	{  SState *ss = (SState *)(s->state);

	   if (ss->Rprefix && ss->Rprefix != ss->Lprefix)
	     /* erase old prefix, in case it's sensitive */
	     { if (ss->RprefixL) bzero((char *)ss->Rprefix,ss->RprefixL); xFree((VOID *)ss->Rprefix); }
	   if (len > 0)
	   {  ss->RprefixL = len; ss->Rprefix = (unsigned char *)xMalloc(len);
	      bcopy(buf,(char *)ss->Rprefix,len);  }
	   else { ss->RprefixL = 0; ss->Rprefix = 0; }
	   return 0;  }

      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((SState *)(s->state))->flags |= IP_PSHDR_FLAG;
	    xTrace0(hmacp, TR_EVENTS,
	       "hmacControlSessn: Turning on IPpseudoheader length fixup.");  }
	    if (!s->numdown) return 0;
      /* **** FALL THROUGH TO DEFAULT CASE, so lower sessions see it too. **** */

      default:
        if (buf && len) bzero(buf, len);   /* prevent info going down */
	if (s->numdown)
	  return xControl(xGetDown(s,TRANSPORT_PROTOCOL), op, buf, len);
	else return -1;
    }
}


static int
hmacControlProtl( self, op, buf, len )
    XObj	self;
    int 	op, len;
    char 	*buf;
{
    xAssert(xIsProtocol(self));

    switch (op) {
      case HASH_SIZE:
	return MD5_SIZE;

      case GETMAXPACKET:
      case GETOPTPACKET:
        {
            int max;
	    if (len < sizeof(max) || !buf) return 0; 
	    if (xControl(xGetDown(self, TRANSPORT_PROTOCOL), op, (char *) &max, sizeof(max)) <= 0) {
		return 0; 
	    }
	    max -= 16; if (max<0) return 0;
            *(int *)buf = max;
            return(sizeof(max));
	}

      case IP_PSEUDOHDR:   /* Turn on flag saying to fixup length on IPpseudoheader. */
	{   ((PState *)(self->state))->flags |= IP_PSHDR_FLAG;
	    xTrace0(hmacp, TR_EVENTS,
	       "hmacControlProtl: Turning on IPpseudoheader length fixup.");  }
      /* **** FALL THROUGH TO DEFAULT CASE, so lower protocols see it too. **** */

      default:
	if (self->numdown) {
	  if (buf && len) bzero(buf, len);   /* prevent info going down */
	  return xControl(xGetDown(self, TRANSPORT_PROTOCOL), op, buf, len);
	}
	else return -1;
  }
  
}


static void
hmacSessnInit(s)
    XObj s;
{
  s->push = hmacPush;
  s->pop = hmacPop;
  s->close = hmacClose;
  s->control = hmacControlSessn;
}


