/*
 * $RCSfile: ppp_fsm.h,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: ppp_fsm.h,v $
 * Revision 1.2  1994/10/27 17:27:28  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:44:34  andrey
 * Initial revision
 *
 *
 * $Revision: 1.2 $
 * $Date: 1994/10/27 17:27:28 $
 */

/*
 * finite state automaton with actions
 *         for the Point-to-Point
 *	    Link Control Protocol, per RFC 1548
 *	    and IP Control Protocol, per RFC 1332
 */

/*
 * implementation configuration params
 * timeouts, retries, etc
 */
typedef struct {
  /* in micro-seconds  */
  u_long timeval;

  /* number of PPP_TERM_REQ to send without reply */
  u_short term_retry;

  /* number of PPP_CONF_REQ to send before we assume peer is croaked */
  u_short config_retry;

  /* number of PPP_CONF_NAK/REJ we receive before we conclude */
  /* that negotiation is impossible */
  u_short config_naks;

} PPP_config;

#define SECONDS 1000000         /* useconds -> seconds */

/*
 * default values for above, from RFC 1548
 */
#define PPP_config_init(PPP_config_ptr)\
   ((PPP_config_ptr)->timeval=3*SECONDS,(PPP_config_ptr)->term_retry=2,\
    (PPP_config_ptr)->config_retry=10,(PPP_config_ptr)->config_naks=10)


/* states */

typedef enum { F_initial=0,
               F_starting,
               F_closed,
               F_stopped,
               F_closing,
               F_stopping,
               F_req_sent,
               F_ack_rcvd,
               F_ack_sent,
               F_opened,
               F_err
} fsm_state;

/*   events	PPP token			RFC1548 name 	*/

typedef enum { F_up=0,			/* UP */
               F_down,			/* DOWN */
               F_open,			/* OPEN */
               F_close,			/* CLOSE */
               F_timeout_pos,		/* TO+ */
               F_timeout_zero,		/* TO- */
               F_rcv_cfg_req_good,	/* RCR+ */
               F_rcv_cfg_req_bad,	/* RCR- */
               F_rcv_cfg_ack,		/* RCA */
               F_rcv_cfg_nak,		/* RCN */
               F_rcv_term_req,		/* RTR */
               F_rcv_term_ack,		/* RTA */
               F_rcv_unk_code,		/* RUC */
               F_rcv_cod_rej_perm,	/* RXJ+ */
               F_rcv_cod_rej_catas, 	/* RXJ- */
               F_rcv_echo_req		/* RXR */
} fsm_event;

#define	NO_EVENT	-1
#define	last_state	F_err
#define last_event	F_rcv_echo_req

/*
 * next state table (events x states)
 */

typedef fsm_state fsm_next_state[last_event+1][last_state];

extern char * event_name[];

struct FSM_info;

typedef	void (*fsm_act)(struct FSM_info *);

/*
 * action table (events x states)
 */
typedef fsm_act fsm_action[last_event+1][last_state];

/*
 * attribute table for state - TIMER_ON if timer is running
 *   during state, else TIMER_OFF
 */
#define	TIMER_ON	1
#define	TIMER_OFF	0

extern unsigned char state_attrs[];
extern char * state_name[];


typedef struct FSM_info {
  Event timerev;                /* timer is running */
  Semaphore Nego_sem;           /* wait for negotiation to be done */
  char waiting_on_nego;         /* way to indicate that call for negotiation */
                                /*  is blocked on Nego_sem above */
  char nego_ok;			/* 1 if went ok, 0 else */

  u_char nak_not_rej;		/* 1 if send nak in scn, 0 if send rej in scn */

  fsm_state state;

  PPP_config * time_options;	/* timeout options */

  Msg * msg;                    /* msg from lcpPop, to communicate with PPP */

  u_short restart_counter;

  fsm_next_state * state_table;
  fsm_action * action_table;

  XObj session;			/* session associated with this table */

  /* callbacks to session */

  xkern_return_t (*process_message)(XObj , Msg * );
  Msg * (*make_config_req_msg)(XObj );
  Msg * (*make_config_ack_msg)(XObj );
  Msg * (*make_config_nak_msg)(XObj );
  Msg * (*make_config_rej_msg)(XObj );
  void (*signal_negotiation_done)(XObj );

} FSM_info;

#define nul     0
void ppp_tlu(FSM_info *);		/* this layer up */
void ppp_tld(FSM_info *);		/* this layer down */
void ppp_tls(FSM_info *);		/* this layer started */
void ppp_tlf(FSM_info *);		/* this layer finished */
void ppp_irc(FSM_info *);		/* initialize restart counter */
void ppp_zrc(FSM_info *);		/* zero restart counter */
void ppp_scr(FSM_info *);		/* send configure request */
void ppp_sca(FSM_info *);		/* send configure ack */
void ppp_scn(FSM_info *);		/* send configure nak */
void ppp_str(FSM_info *);		/* send terminate request */
void ppp_sta(FSM_info *);		/* send terminate ack */
void ppp_scj(FSM_info *);		/* send code reject */
void ppp_ser(FSM_info *);		/* send echo reply */

void ppp_irc_scr(FSM_info *);		/* irc, then scr */
void ppp_irc_scr_sca(FSM_info *);	/* irc, then scr, then sca */
void ppp_irc_scr_scn(FSM_info *);	/* irc, then scr, then scn */
void ppp_irc_str(FSM_info *);		/* et cetera */
void ppp_tld_irc_str(FSM_info *);
void ppp_tld_scr_sca(FSM_info *);
void ppp_tld_scr_scn(FSM_info *);
void ppp_irc_tlu(FSM_info *);
void ppp_tld_scr(FSM_info *);
void ppp_tld_zrc_sta(FSM_info *);
void ppp_sca_tlu(FSM_info *);
void ppp_error(FSM_info *);

void ppp_process_event(FSM_info * , fsm_event);	/* crank automaton on event */

int ppp_decode_packet(Msg * msg, int * id, int * len);

