/*
 * $RCSfile: watchdog.c,v $

 * This WATCHDOG protocol implements a simple watchdog timer.  The
 * duraction of the timer is set via a rom option and the timer itself
 * is reset via a WATCHDOG control operation.  It is not possible to
 * either demux or push into the WATCHDOG protocol.

 * When the timer expires the xKernel is shutdown by calling
 * xRapture().  This will nicely shutdown all protocols.

 * This protocol can be used w/o reseting the watchdog timer to insure
 * an xKernel shutsdown after a specified time.  This is quite useful
 * when running the xKernel on the physical ethernet of a remote
 * machine.  I.e. it allows testing w/o being present at the system's
 * console.
 
 * Copyright (c) 1996	Arizona Board of Regents
 *
 * $Log: watchdog.c,v $
 * Revision 1.2  1997/03/10 22:11:39  mjk
 * *** empty log message ***
 *
 * Revision 1.1  1996/11/04 23:36:06  mjk
 * Initial revision
 *
 *
 */


#include "xkernel.h"
#include "watchdog.h"


void			watchdog_init(XObj);
static xkern_return_t	watchdog_openenable(XObj, XObj, XObj, Part *);
static xkern_return_t	watchdog_shutdown(XObj);
static int		watchdog_control(XObj, int, char *, int);
static xkern_return_t   watchdog_rom_timeout(XObj, char **, int, int, void *);
static void		watchdog_timer(Event e, void *arg);


/* 
 * Typedefs 
 */

typedef struct {		/* Protocol State 			*/
  unsigned	timeout;		/* timeout length (seconds)	*/
  unsigned	resets;			/* number of resets		*/
  int		shutdown;		/* shutdown predicate 		*/
} PState;



/* 
 * Globals 
 */

int	tracewatchdogp;

					/* rom options			*/
static XObjRomOpt	watchdog_opts[] = {
  { "timeout",	3, watchdog_rom_timeout }
};



/* 
 *  Public Functions
 */

void
watchdog_init(XObj self) 
{
  PState	*ps;

  xTraceP0(self, TR_GROSS_EVENTS, "init");

  ps = X_NEW(PState);
  ps->timeout	= 0;
  ps->shutdown	= 0;
  self->state = ps;

  findXObjRomOpts(self, watchdog_opts,
		  (sizeof watchdog_opts) / sizeof (XObjRomOpt), NULL);

				/* start watchdog timer 		*/
  if ( ps->timeout ) {
    evDetach(evSchedule(watchdog_timer, self, 0));
  }
  
  self->openenable  = watchdog_openenable;
  self->shutdown    = watchdog_shutdown;
  self->control	    = watchdog_control;
} /* watchdog_init */



/* 
 * Public Protocol Methods 
 */

static xkern_return_t
watchdog_openenable(XObj self, XObj hlpRcv, XObj hlpType, Part *p)
{
  return XK_SUCCESS;
} /* watchdog_openenable */


static xkern_return_t
watchdog_shutdown(XObj self) 
{
  PState		*ps = self->state;

  xTraceP0(self, TR_MAJOR_EVENTS, "shutdown");
  ps->shutdown = 1;		/* shutdown timer event			*/

  return XK_SUCCESS;
} /* watchdog_shutdown */


static int
watchdog_control(XObj self, int opcode, char *buf, int buflen)
{
  PState	*ps = (PState *)self->state;

  switch ( opcode ) {
    case WATCHDOG:
      ps->resets++;
      return 0;
    default:
      if ( xGetDown(self, 0) )
	return xControl(xGetDown(self, 0), opcode, buf, buflen);
      else
	return -1;
  }
} /* watchdog_control */



/* 
 * Private Protocol Methods 
 */

static xkern_return_t 
watchdog_rom_timeout(XObj self, char **argv, int argc, int line, void *arg)
{
  PState	 *ps = (PState *)self->state;

  if ( argc == 3 &&
       sscanf(argv[2], "%d", &ps->timeout) == 1 )
  {
    ps->resets  = 0;
    xTraceP1(self, TR_MAJOR_EVENTS, "timer set at %d seconds",
	     ps->timeout);
  }
  else {
    return XK_FAILURE;
  }

  return XK_SUCCESS;
} /* ethkern_rom_ipat */


static void
watchdog_timer(Event e, void *arg)
{
  XObj		self = (XObj)arg;
  PState	*ps = (PState *)self->state;

  ps->resets = 0;
  while ( ! ps->shutdown ) {
    xTraceP0(self, TR_EVENTS, "sleep");
    Delay( ps->timeout * 1000);
    xTraceP0(self, TR_EVENTS, "wakeup");
    if ( ! ps->resets ) {
      xTraceP0(self, TR_MAJOR_EVENTS, "watchdog expired");
      xRapture();		/* shutdown all protocols 		*/
      exit(0);			/* exit process 			*/
    }
  }
} /* watchdog_timer */



/* watchdog.c ends here */
 



