/*
 * sarCtlr.h
 * Include file for SAR interface board driver
 *
 * $RCSfile: sarCtlr.h,v $
 *
 * HISTORY
 * $Log: sarCtlr.h,v $
 * Revision 1.2  1994/06/04 05:06:21  druschel
 * Cleanup
 *
 * Revision 1.1  1994/05/22  22:28:07  druschel
 * Initial revision
 *
 *
 */

/* Definitions specific to the SAR driver */
/* CSR Definitions */

#define SAR_CSR_MASK 0x000000ff
/* 
  xmit side: xmitter is idle, ready
  rcv side: receiver is ready
*/
#define SAR_RDY   0x00000001 

/* Indicate error condition occured */
#define SAR_ERROR 0x00000002

/* reset */
#define SAR_RESET 0x00000004 

#define SAR_TEST 0x00000008

/*
 * rcv side: a packet was queued
 */
#define SAR_PRDY  0x00000020

/*
 * rcv side: need buffers
 */
#define SAR_NEEDB  0x00000010

/* 
 * This bit is used to tell the interrupt handler which SAR, the xmit
 * or the receive, caused the interrupt. 
 */
#define SAR_INTR 0x00000080



/* Software control structure definitions */
#define SAR_OPEN 1
#define SAR_CLOSED 0
#define SAR_LOCKED 1
#define SAR_UNLOCKED 0
/* Define ROM offsets for TurboChannel */
#define ROM_WIDTH 0x3E0 /* offset for rom_width field */
#define ROM_STRIDE 0x3E4 /* offset for rom_stride field */
#define ROM_SIZE 0x3E8 /* offset for rom_size field */
#define ROM_SLOT 0x3EC /* offset for rom slot_size field */
#define ROM_PAT1 0x3F0 /* offset for pattern 1 field, pattern = 0x55 */
#define ROM_PAT2 0x3F4 /* offset for pattern 2 field, pattern = 0x00 */
#define ROM_PAT3 0x3F8 /* offset for pattern 3 field, pattern = 0xAA */
#define ROM_PAT4 0x3FC /* offset for pattern 4 field, pattern = 0xFF */

/* Random */
#define NSARD 1 /* Number of SARs */
#define MAX_RETRY 100000 /*Max number of retrys for spin loops */
#define SARPRI PZERO /* priority of sar when it sleeps */
#define DPMEM_SIZE 32768 /* size of Dual-ported memory 32K words*/

#define SARMTU 16368 /* define SAR's max transmission unit (372 * 44) */

#ifdef SIM_RCV_HACK
#undef SARMTU
#define SARMTU 16384
/*#define SARMTU 32688*/
#endif

/* cell payload length */
#define MAX_CELL_LEN 44

/* cell seqNum range (0-1023) */
#define CELL_SEQNUM_RANGE 1024

/* top bit of VCI indicates last fragment */
#define SAR_LAST_FRAG 0x80000000


typedef struct sar_buf {
  u_int physAddr;    /* base physical address for data */
  int     count;     /* byte count */
} SarBuf;


typedef struct sar_reg {

  char eprom[65536] ; 

  /****************************************/
  /* Control/status for Write half of SAR */
  /****************************************/

  /* XMIT_QUEUE_SIZE must be a power of 2! */
#define XMIT_QUEUE_SIZE 16

  volatile int xmitCSR;   /* Control/status register for write*/

  /* xmit queue, written by host, read by board 
     Invariant:
     (1) (xmitQHead + 1) % XMIT_QUEUE_SIZE == xmitQTail => queue is full
     (2) xmitQHead == xmitQTail => queue is empty
     (3) !empty => xmitQueue[xmitQTail] is next buffer to xmit
     (4) xmitQueue[xmitQHead] is next free descriptor
   */
  struct xmit_queue {
    u_int physAddr; /* physical address of data in host memory */
    int   count;    /* Byte count for write*/
    int   vci;      /* VCI for write */
    int   seqNum;   /* next cell seqNum for this VCI */
  } xmitQueue[XMIT_QUEUE_SIZE];
  
  int xmitQHead;          /* xmit queue head, written by host, read by board */
  volatile int xmitQTail; /* xmit queue tail, read by host, written by board */

  /* statistics */
  volatile int xmitIdles;
  volatile int xmitFull;
  volatile int xmitEmpty;

  /* fake local eth address */
  ETHhost ethAddr;
 
  /* 
   * memory space left is 64K - bytes used above 
   */
  char  w_space[65536 - 6 * sizeof(int) - sizeof(ETHhost) -
		XMIT_QUEUE_SIZE * sizeof(struct xmit_queue)] ; 


  /****************************************/
  /* Control/status for Read half of SAR  */
  /****************************************/

  volatile int rcvCSR;      /* control/status register */

  /* RCV_QUEUE_SIZE must be a power of 2! */
#define RCV_QUEUE_SIZE 64
  /* rcv queue
     used to pass packets (or fragments thereof) to the host
     Invariant:
     (1) (rcvQHead + 1) % RCV_QUEUE_SIZE == rcvQTail => queue is full
     (2) rcvQHead == rcvQTail => queue is empty
     (3) !empty => rcvQueue[rcvQTail] is next received fragment descriptor
     (4) rcvQueue[rcvQHead] is next free slot
   */
  volatile struct rcv_queue {
    int   bufIndex;  /* buffer Index */
    int   count;     /* length of received fragment */
    int   vci;       /* VCI of received fragment */
  } rcvQueue[RCV_QUEUE_SIZE];
  
  volatile int rcvQHead; /* rcv queue head, written by board, read by host */
  int rcvQTail;          /* rcv queue tail, read by board, written by host */


  /* FREEB_QUEUE_SIZE must be a power of 2! */
#define FREEB_QUEUE_SIZE 64
  /* free buffer queue
     used to supply free buffers to the rcv controller  
     Invariant:
     (1) (freeBQHead + 1) % FREEB_QUEUE_SIZE == freeBQTail => queue is full
     (2) freeBQHead == freeBQTail => queue is empty
     (3) !empty => freeBQueue[freeBQTail] is next free buffer
     (4) freeBQueue[freeBQHead] is next free buffer descriptor
   */
  struct freeb_queue {

    u_int physAddr; /* Physical address of buffer in host memory */
    int   bufIndex; /* buffer Index (meaningful only to host) */
    int   size;     /* size of buffer */
  } freeBQueue[FREEB_QUEUE_SIZE];
  
  int freeBQHead;          /* queue head, written by host, read by board */
  volatile int freeBQTail; /* queue tail, read by host, written by board */

  /* bad cell dump */
  volatile char rcvDump[MAX_CELL_LEN];

  /* statistics */
  volatile int rcvCRCErrors;        /* cells with header CRC error */
  volatile int rcvBQueueUnderflows; /* free buffer underflows */
  volatile int rcvQueueOverflows;   /* receive queue overflows */
  volatile int rcvOutOfOrderCells;  /* out-of-order arrivals, cells */
  volatile int rcvOutOfOrderFrags;  /* out-of-order arrivals, frags */
  volatile int rcvIdles;            /* emptied the header FIFO */
  volatile int rcvHdrFIFOFull;      /* header FIFO seen full */
  volatile int rcvDMAFIFOFull;      /* DMA FIFO seen full */

  u_int rcvDumpAddr;                /* phys. host addr to dump cells */

  /* 
   * memory space left is 64K - space used above 
   */
  char r_data[65536 - 14 * sizeof(int) -
	      RCV_QUEUE_SIZE * sizeof(struct rcv_queue) - 
	      FREEB_QUEUE_SIZE * sizeof(struct freeb_queue) - 
	      MAX_CELL_LEN * sizeof(char)];

  /****************************************/
  /* Semaphore for Dual-ported memory     */
  /****************************************/

  char unused1[65536];
  volatile int xmitSema ; /* xmit side dual-port memory semaphore */
  char unused2[65536 - sizeof(int)];
  volatile int rcvSema ; /* receive side dual-port memory semaphore */

} SarReg ;



typedef struct sar_softc {
  Scout_IPL my_ipl;     /* osiris int. priority */
  SarReg *sc_regs;      /* a pointer to the SarReg structure */
  int sc_status;        /* retains number of processes that have opened sar */
  int sc_unit;          /* unit number of device */
  char sc_devname[20] ; /* device name */

#ifdef IPC_FBUF
  struct mna *allocator; /* msg (fbuf) allocator */
#endif

  Scout_Thread *rcvThread;   /* receiver thread (only one for now) */

  /*
   * descriptors for free receive buffers
   */
  struct freeb_desc {
    Msg   msg;
    char* virtAddr;
    u_int physAddr;
  } rcvFreeBuf[FREEB_QUEUE_SIZE];
  
  /* saved copies of transmitted messages */
  Msg xmitMsg[XMIT_QUEUE_SIZE];

  ETHhdr ethHdr;	/* fake eth hdr for incoming packets */

} SarSoftc;
